#include "perfmon.h"
#include "stdlib.h"

/*

The encoding of mov crx, reg instruction:

MOV  Move to/from Control Registers
CR0 from register         0000 1111 : 0010 0010 : 11 000 reg
CR2 from register         0000 1111 : 0010 0010 : 11 010 reg
CR3 from register         0000 1111 : 0010 0010 : 11 011 reg
CR4 from register         0000 1111 : 0010 0010 : 11 100 reg
register from CR0-CR4     0000 1111 : 0010 0000 : 11 eee reg

MOV Move to/from Debug Registers
DR0-DR3 from register     0000 1111 : 0010 0011 : 11 eee reg
DR4-DR5 from register     0000 1111 : 0010 0011 : 11 eee reg
DR6-DR7 from register     0000 1111 : 0010 0011 : 11 eee reg
register from DR6-DR7     0000 1111 : 0010 0001 : 11 eee reg
register from DR4-DR5     0000 1111 : 0010 0001 : 11 eee reg
register from DR0-DR3     0000 1111 : 0010 0001 : 11 eee reg

Encoding of Special-Purpose Register (eee) Field
eee         Control Register        Debug Register
000         CR0                     DR0
001         Reserved*               DR1
010         CR2                     DR2
011         CR3                     DR3
100         CR4                     Reserved*
101         Reserved*               Reserved*
110         Reserved*               DR6
111         Reserved*               DR7

Encoding of reg Field When w Field is Not Present in Instruction
reg Field   Register Selected during    Register Selected during
            16-Bit Data Operations      32-Bit Data Operations

000         AX                          EAX
001         CX                          ECX
010         DX                          EDX
011         BX                          EBX
100         SP                          ESP
101         BP                          EBP
110         SI                          ESI
111         DI                          EDI

*/

ULONG GetCR4() {
  ULONG retV;
  __asm {
    push eax;
    // mov eax, cr4
    //                   eee reg
    // -> 0x0F 0x20  %11 100 000  
    __asm _emit 0x0F
    __asm _emit 0x20
    __asm _emit 0xE0
    mov retV, eax;
    pop eax;
  }
  return retV;
}

void SetCR4(ULONG val) {
  __asm {
    push eax;
    mov eax, val;
    // mov cr4, eax
    //                   eee reg
    // -> 0x0F 0x22  %11 100 000  
    __asm _emit 0x0F
    __asm _emit 0x22
    __asm _emit 0xE0  
    pop eax;
  }
}

NTSTATUS DriverEntry( IN PDRIVER_OBJECT  DriverObject, IN PUNICODE_STRING RegistryPath) {
/*++
Routine Description:
    This routine is the entry point for the driver.  It is responsible
    for setting the dispatch entry points in the driver object and creating
    the device object.  Any resources such as ports, interrupts and DMA
    channels used must be reported.  A symbolic link must be created between
    the device name and an entry in \DosDevices in order to allow Win32
    applications to open the device.

Arguments:

    DriverObject - Pointer to driver object created by the system.

Return Value:

    STATUS_SUCCESS if the driver initialized correctly, otherwise an error
    indicating the reason for failure.

--*/
    PLOCAL_DEVICE_INFO pLocalInfo;  // Device extension:
                                    //      local information for each device.
    UNICODE_STRING unicodePrompt;
    NTSTATUS Status;
    PDEVICE_OBJECT DeviceObject;
    CM_RESOURCE_LIST ResourceList;  // Resource usage list to report to system




    // Initialize the driver object dispatch table.
    // NT sends requests to these routines.

    DriverObject->MajorFunction[IRP_MJ_CREATE]          = GpdDispatch;
    DriverObject->MajorFunction[IRP_MJ_CLOSE]           = GpdDispatch;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]  = GpdDispatch;
    DriverObject->DriverUnload                          = GpdUnload;

    // Create our device.
    Status = GpdCreateDevice(
                    VDTFG_DEVICE_NAME,
                    PERFMON_TYPE,
                    DriverObject,
                    &DeviceObject
                    );

    if ( NT_SUCCESS(Status) ) {

         // Initialize the local driver info for each device object.
        pLocalInfo = (PLOCAL_DEVICE_INFO)DeviceObject->DeviceExtension;
        pLocalInfo->DeviceObject    = DeviceObject;
        pLocalInfo->DeviceType      = PERFMON_TYPE;

    }

    return Status;
}

NTSTATUS GpdCreateDevice(IN PWSTR PrototypeName, IN DEVICE_TYPE DeviceType, IN  PDRIVER_OBJECT DriverObject, OUT PDEVICE_OBJECT *ppDevObj)

/*++

Routine Description:
    This routine creates the device object and the symbolic link in
    \DosDevices.

    Ideally a name derived from a "Prototype", with a number appended at
    the end should be used.  For simplicity, just use the fixed name defined
    in the include file.  This means that only one device can be created.

    A symbolic link must be created between the device name and an entry
    in \DosDevices in order to allow Win32 applications to open the device.

Arguments:

    PrototypeName - Name base, # WOULD be appended to this.
    DeviceType - Type of device to create
    DriverObject - Pointer to driver object created by the system.
    ppDevObj - Pointer to place to store pointer to created device object

Return Value:

    STATUS_SUCCESS if the device and link are created correctly, otherwise
    an error indicating the reason for failure.

--*/


{
    NTSTATUS Status;                        // Status of utility calls
    UNICODE_STRING NtDeviceName;
    UNICODE_STRING Win32DeviceName;


    // Get UNICODE name for device.

    RtlInitUnicodeString(&NtDeviceName, PrototypeName);

    Status = IoCreateDevice(                             // Create it.
                    DriverObject,
                    sizeof(LOCAL_DEVICE_INFO),
                    &NtDeviceName,
                    DeviceType,
                    0,
                    FALSE,                      // Not Exclusive
                    ppDevObj
                    );

    if (!NT_SUCCESS(Status))
        return Status;             // Give up if create failed.

    // Clear local device info memory
    RtlZeroMemory((*ppDevObj)->DeviceExtension, sizeof(LOCAL_DEVICE_INFO));

    //
    // Set up the rest of the device info
    //  These are used for IRP_MJ_READ and IRP_MJ_WRITE which we don't use
    //
    //  (*ppDevObj)->Flags |= DO_BUFFERED_IO;
    //  (*ppDevObj)->AlignmentRequirement = FILE_BYTE_ALIGNMENT;
    //

    RtlInitUnicodeString(&Win32DeviceName, DOS_DEVICE_NAME);

    Status = IoCreateSymbolicLink( &Win32DeviceName, &NtDeviceName );

    if (!NT_SUCCESS(Status)) {     // If we we couldn't create the link then
        IoDeleteDevice(*ppDevObj); //  abort installation.
    }

    return Status;
}


NTSTATUS GpdDispatch(IN    PDEVICE_OBJECT pDO, IN    PIRP pIrp)

/*++

Routine Description:
    This routine is the dispatch handler for the driver.  It is responsible
    for processing the IRPs.

Arguments:

    pDO - Pointer to device object.

    pIrp - Pointer to the current IRP.

Return Value:

    STATUS_SUCCESS if the IRP was processed successfully, otherwise an error
    indicating the reason for failure.

--*/

{
    PLOCAL_DEVICE_INFO pLDI;
    PIO_STACK_LOCATION pIrpStack;
    NTSTATUS Status;

    //  Initialize the irp info field.
    //      This is used to return the number of bytes transfered.

    pIrp->IoStatus.Information = 0;
    pLDI = (PLOCAL_DEVICE_INFO)pDO->DeviceExtension;    // Get local info struct

    pIrpStack = IoGetCurrentIrpStackLocation(pIrp);

    //  Set default return status
    Status = STATUS_NOT_IMPLEMENTED;

    // Dispatch based on major fcn code.

    switch (pIrpStack->MajorFunction) {
        case IRP_MJ_CREATE:
        case IRP_MJ_CLOSE:
            // We don't need any special processing on open/close so we'll
            // just return success.
            Status = STATUS_SUCCESS;
            break;

        case IRP_MJ_DEVICE_CONTROL:
            //  Dispatch on IOCTL
            switch (pIrpStack->Parameters.DeviceIoControl.IoControlCode) {

            case IOCTL_WRITE_MSR:
                Status = IoctlWriteMSR(
                            pLDI,
                            pIrp,
                            pIrpStack,
                            pIrpStack->Parameters.DeviceIoControl.IoControlCode);
                break;

            case IOCTL_READ_MSR:
                Status = IoctlReadMSR(
                            pLDI,
                            pIrp,
                            pIrpStack,
                            pIrpStack->Parameters.DeviceIoControl.IoControlCode);
                break;

            case IOCTL_ENABLE_CR4_BITS:
                Status = IoctlSetCR4(
                            pLDI,
                            pIrp,
                            pIrpStack,
                            pIrpStack->Parameters.DeviceIoControl.IoControlCode);
                break;



            }
            break;
    }

    // We're done with I/O request.  Record the status of the I/O action.
    pIrp->IoStatus.Status = Status;
    // Don't boost priority when returning since this took little time.
    IoCompleteRequest(pIrp, IO_SOUND_INCREMENT );
    return Status;
}

NTSTATUS IoctlWriteMSR(IN PLOCAL_DEVICE_INFO pLDI, IN PIRP pIrp, IN PIO_STACK_LOCATION IrpStack, IN ULONG IoctlCode)

/*++

Routine Description:
    This routine processes the IOCTLs which write to the MSR:s

Arguments:

    pLDI        - our local device data
    pIrp        - IO request packet
    IrpStack    - The current stack location
    IoctlCode   - The ioctl code from the IRP

Return Value:
    STATUS_SUCCESS           -- OK

    STATUS_INVALID_PARAMETER -- The buffer sent to the driver
                                was too small to contain the
                                port, or the buffer which
                                would be sent back to the driver
                                was not a multiple of the data size.

    STATUS_ACCESS_VIOLATION  -- An illegal port number was given.

--*/

{
                                // NOTE:  Use METHOD_BUFFERED ioctls.
    PULONG pIOBuffer;           // Pointer to transfer buffer
                                //      (treated as array of longs).
    ULONG InBufferSize ;        // Amount of data avail. from caller.
    ULONG OutBufferSize ;       // Max data that caller can accept.    

    ULONG   RegNumber; 
    ULONG   LoData;
    ULONG   HiData;

    // Size of buffer containing data from application
    InBufferSize  = IrpStack->Parameters.DeviceIoControl.InputBufferLength;

    // Size of buffer for data to be sent to application
    OutBufferSize = IrpStack->Parameters.DeviceIoControl.OutputBufferLength;

    // NT copies inbuf here before entry and copies this to outbuf after return,
    // for METHOD_BUFFERED IOCTL's.
    pIOBuffer     = (PULONG) pIrp->AssociatedIrp.SystemBuffer;
    
    pIrp->IoStatus.Information = 0;
  
    if ( InBufferSize < 3  * sizeof(ULONG) ) {
        return STATUS_INVALID_PARAMETER;
    }

    RegNumber = *pIOBuffer++;
    LoData = *pIOBuffer++;
    HiData = *pIOBuffer++;

    
    __asm {
      push edx;
      push eax;
      push ecx;
      mov ecx, RegNumber;
      mov eax, LoData;
      mov edx, HiData;
      // do the WRMSR instruction
      _asm _emit 0x0F
      _asm _emit 0x30
      pop ecx;
      pop eax;
      pop edx;
    }
    
    return STATUS_SUCCESS;
}


NTSTATUS IoctlReadMSR(IN PLOCAL_DEVICE_INFO pLDI, IN PIRP pIrp, IN PIO_STACK_LOCATION IrpStack, IN ULONG IoctlCode)

/*++

Routine Description:
    This routine processes the IOCTLs which write to the MSR:s

Arguments:

    pLDI        - our local device data
    pIrp        - IO request packet
    IrpStack    - The current stack location
    IoctlCode   - The ioctl code from the IRP

Return Value:
    STATUS_SUCCESS           -- OK

    STATUS_INVALID_PARAMETER -- The buffer sent to the driver
                                was too small to contain the
                                port, or the buffer which
                                would be sent back to the driver
                                was not a multiple of the data size.

    STATUS_ACCESS_VIOLATION  -- An illegal port number was given.

--*/

{
                                // NOTE:  Use METHOD_BUFFERED ioctls.
    PULONG pIOBuffer;           // Pointer to transfer buffer
                                //      (treated as array of longs).
    ULONG InBufferSize ;        // Amount of data avail. from caller.
    ULONG OutBufferSize ;       // Max data that caller can accept.    

    ULONG   RegNumber; 
    ULONG   LoData;
    ULONG   HiData;

    // Size of buffer containing data from application
    InBufferSize  = IrpStack->Parameters.DeviceIoControl.InputBufferLength;

    // Size of buffer for data to be sent to application
    OutBufferSize = IrpStack->Parameters.DeviceIoControl.OutputBufferLength;

    // NT copies inbuf here before entry and copies this to outbuf after return,
    // for METHOD_BUFFERED IOCTL's.
    pIOBuffer     = (PULONG) pIrp->AssociatedIrp.SystemBuffer;
    
    pIrp->IoStatus.Information = 0;
  
    if ( InBufferSize < sizeof(ULONG) ) {
        return STATUS_INVALID_PARAMETER;
    }

    RegNumber = *pIOBuffer;    
    
    __asm {
      push edx;
      push eax;
      push ecx;
      mov ecx, RegNumber;      
      // do the RDMSR instruction
      _asm _emit 0x0F
      _asm _emit 0x32
      mov LoData, eax;
      mov HiData, edx;
      pop ecx;
      pop eax;
      pop edx;
    }

    *pIOBuffer++ = LoData;
    *pIOBuffer++ = HiData;    
    pIrp->IoStatus.Information = 8;
    
    return STATUS_SUCCESS;
}


NTSTATUS IoctlSetCR4(IN PLOCAL_DEVICE_INFO pLDI, IN PIRP pIrp, IN PIO_STACK_LOCATION IrpStack, IN ULONG IoctlCode)

/*++

Routine Description:
    This routine processes the IOCTLs which read and write the CR4 register

Arguments:

    pLDI        - our local device data
    pIrp        - IO request packet
    IrpStack    - The current stack location
    IoctlCode   - The ioctl code from the IRP

Return Value:
    STATUS_SUCCESS           -- OK

    STATUS_INVALID_PARAMETER -- The buffer sent to the driver
                                was too small to contain the
                                port, or the buffer which
                                would be sent back to the driver
                                was not a multiple of the data size.

    STATUS_ACCESS_VIOLATION  -- An illegal port number was given.

--*/

{
                                // NOTE:  Use METHOD_BUFFERED ioctls.
    PULONG pIOBuffer;           // Pointer to transfer buffer
                                //      (treated as array of longs).
    ULONG InBufferSize ;        // Amount of data avail. from caller.
    ULONG OutBufferSize ;       // Max data that caller can accept.    

    ULONG   SetBits; 
    ULONG   PrevCR4;
    ULONG   LoData;
    ULONG   HiData;

    // Size of buffer containing data from application
    InBufferSize  = IrpStack->Parameters.DeviceIoControl.InputBufferLength;

    // Size of buffer for data to be sent to application
    OutBufferSize = IrpStack->Parameters.DeviceIoControl.OutputBufferLength;

    // NT copies inbuf here before entry and copies this to outbuf after return,
    // for METHOD_BUFFERED IOCTL's.
    pIOBuffer     = (PULONG) pIrp->AssociatedIrp.SystemBuffer;
    
    pIrp->IoStatus.Information = 0;
  
    if ( InBufferSize < sizeof(ULONG) || OutBufferSize < sizeof(ULONG)) {
        return STATUS_INVALID_PARAMETER;
    }

    SetBits = *pIOBuffer;    
    SetBits &= 0x104; // mask only TSD and PCE bits from the input
    
    PrevCR4 = GetCR4();        
    *pIOBuffer = PrevCR4;
    PrevCR4 |= SetBits;
    PrevCR4 &= (SetBits | ~0x104);
    SetCR4(PrevCR4);
    
    pIrp->IoStatus.Information = 4;
    
    return STATUS_SUCCESS;
}



VOID GpdUnload(PDRIVER_OBJECT DriverObject)

/*++

Routine Description:
    This routine prepares our driver to be unloaded.  It is responsible
    for freeing all resources allocated by DriverEntry as well as any
    allocated while the driver was running.  The symbolic link must be
    deleted as well.

Arguments:

    DriverObject - Pointer to driver object created by the system.

Return Value:

    None

--*/

{
    PLOCAL_DEVICE_INFO pLDI;
    CM_RESOURCE_LIST NullResourceList;
    BOOLEAN ResourceConflict;
    UNICODE_STRING Win32DeviceName;
    // Find our global data
    pLDI = (PLOCAL_DEVICE_INFO)DriverObject->DeviceObject->DeviceExtension;

    // Assume all handles are closed down.
    // Delete the things we allocated - devices, symbolic links
    RtlInitUnicodeString(&Win32DeviceName, DOS_DEVICE_NAME);
    IoDeleteSymbolicLink(&Win32DeviceName);
    IoDeleteDevice(pLDI->DeviceObject);
}


