(*
 * LANGUAGE    : ANS Forth
 * PROJECT     : Forth Environments
 * DESCRIPTION : Access to the Performance Counter library in perfctrs.dll (perfmon.sys)
 * CATEGORY    : Systems Tool
 * AUTHOR      : Marcel Hendrix 
 * LAST CHANGE : December 23, 2000, Marcel Hendrix, interpretation of evt_duration? changed
 * LAST CHANGE : December 18, 2000, Marcel Hendrix, perfsel3 and perfsel4 were set the same.
 * LAST CHANGE : December 14, 2000, Marcel Hendrix,
 *)


        NEEDS -miscutil
	NEEDS -dynlink

        REVISION -perfctrs " Performance counter Version 1.02 "

	PRIVATES

DOC
(* 
 The Performance Monitoring Counters related functions ... 
  1. Only work for Athlon (currently)
  2. Require set up of a device driver

 Set up
 ------
 Copy PERFMON.SYS to %systemroot%\system32\drivers
 Run LOADDRV and install, then start PERFMON.SYS.
 Rebooting is not necessary. The system will remember 
 to start the service at each reboot.

 When something goes wrong, Win2K refuses to boot or 
 reboots spontaneously directly after startup. In that 
 case, start up in "safe mode" and delete perfmon.sys.

 Example (execution time o do-test is 1 second)
 ==============================================
	FORTH> 0 do-test 1.000 seconds elapsed.

	[00C5] retired_far_control_transfers   : 626
	[00C7] retired_resync_branches         : 3
	[00CE] ints_masked_while_pending       : 0
	[00CF] ints_taken                      : 193

	[00CD] ints_masked                     : 33,844
	[1F42] data_cache_refill_l2_all        : 968
	[1F43] data_mem_refs_all               : 279
	[1F44] data_cache_writebacks_all       : 1,279

	[0045] l1_dltb_misses_l2_dltb_hits     : 337
	[0046] l1&l2_dtlb_misses               : 491
	[0047] misaligned_data_references      : 14
	[0080] ifu_ifetch                      : 241,072,624

	[0081] ifu_ifetch_miss                 : 398
	[0084] l1_itlb_misses                  : 36
	[0085] l1&l2_itlb_misses               : 92
	[00C0] retired_instructions            : 1,004,401,986

	[00C1] retired_ops                     : 1,124,990,805
	[00C2] retired_branches                : 40,179,116
	[00C3] retired_mispredicted_branches   : 784
	[00C4] retired_taken_branch_mispredict : 40,177,633

	[0040] data_cache_access               : 401,821,336
	[0041] data_cache_misses               : 1,682
	[1042] data_cache_refill_l2_modified   : 350
	[0842] data_cache_refill_l2_owner      : 0

	[0442] data_cache_refill_l2_exclusive  : 92
	[0242] data_cache_refill_l2_shared     : 1
	[0142] data_cache_refill_l2_invalid    : 212
	[1043] data_mem_refs_modified          : 130

	[0843] data_mem_refs_owner             : 0
	[0443] data_mem_refs_exclusive         : 62
	[0243] data_mem_refs_shared            : 0
	[0143] data_mem_refs_invalid           : 0

	[1044] data_cache_writebacks_modified  : 510
	[0844] data_cache_writebacks_owner     : 0
	[0444] data_cache_writebacks_exclusive : 223
	[0244] data_cache_writebacks_shared    : 0
*)
ENDDOC

0 VALUE perfclib PRIVATE

: UNLOAD-PERFC ( -- )	
	perfclib 0= IF exit ENDIF
	perfclib  CLEAR perfclib library-close 
[ LINUX? ] [IF] ABORT" libperfctrs.so.x :: close failed" 
	 [ELSE]	ABORT" perfctrs.dll :: close failed" 
	 [THEN] ;

-- This assumes PERFCTRS.DLL is in the path
: LOAD-PERFC ( -- )
	perfclib IF exit ENDIF
[ LINUX? ] [IF] S" /lib/libperfctrs.so.1" library-open 
		?DUP IF library-error CR TYPE ABORT" libperfctrs.so.1 :: open failed" ENDIF
	 [ELSE]	S" perfctrs.dll" library-open 
		?DUP IF library-error CR TYPE ABORT" perfctrs.dll :: open failed" ENDIF
	 [THEN]	TO perfclib ;P

: LIB:	( "name" -- )
	LOAD-PERFC  
	BL <WORD> perfclib library-find ABORT" library-find: not in perfctrs"
	ALITERAL ; IMMEDIATE

: ZTYPE ( addr -- )
	BEGIN  C@+ ?DUP
	WHILE  DUP ^J = IF  ^M EMIT  ENDIF 
	       EMIT
	REPEAT DROP ;P

\ Inquire and return the cpu capabilities in a CPU_ID struct.
\ Return true if TimeStamp() can be used (result based on CPUID).
\ Return true if performance monitoring counters can be used (result based on CPUID).

: TimeStamp?	0 LIB: CanDoTimeStamp FOREIGN 0<> ; ( -- bool )
: PerfMon?	0 LIB: CanDoPerfMon   FOREIGN 0<> ; ( -- bool )
: 'GetCPU_ID	0 LIB: GetCPU_ID      FOREIGN ; ( -- addr )
: .PrintID	'GetCPU_ID 1  LIB: PrintID FOREIGN  ZTYPE ; ( -- ) 

\ Sets event counter ctr to count the specified event
\  user? == true -> count events in user mode
\  os? == true -> count events in kernel mode
\  numbers? == true -> count events, not event duration

: SetEventCounter ( ctr event# user? os? numbers? -- ) 
	5 LIB: SetEventCounter VFOREIGN ;

: DisableCounting ( -- )
	0 LIB: DisableEventCounting VFOREIGN ;

\ Read and write model-specific registers.
\ Warning: a non-existent MSR will cause a blue-screen.
\ IAM : byte1 == CPU, byte0 == #counters, -1 : no permission to touch perfcounters
\ Byte1: 4 = Pentium Pro  Byte0 = 2
\        5 = P5			  2
\  	 6 = P6			  2
\        7 = Athlon K7		  4

: InitMSR	0 LIB: InitMSR      FOREIGN ; ( -- IAM )
: ExitMSR	0 LIB: ExitMSR     VFOREIGN ; ( -- )
: TimeStamp	0 LIB: TimeStamp   DFOREIGN ; ( -- ud )
: ReadPerfCtr	1 LIB: ReadPerfCtr DFOREIGN ; ( reg# -- ud | -1. ) 
: WriteMSR	3 LIB: WriteMSR     FOREIGN ; ( reg# ud -- ) \ fails silently
: ReadMSR	1 LIB: ReadMSR     DFOREIGN ; ( reg# -- ud | -1. ) 

\ Magic constants
\ Athlon countable events

: =:=	CREATE 	, @LATEST , ;P
: .EVT	@+ U>D BASE @ >R HEX <#  BL HOLD ']' HOLD # # # # '[' HOLD #> R> BASE !  TYPE
	@ ID$ DUP >R TYPE #32 R> - SPACES DROP ;P

$0040 =:= data_cache_access		
$0041 =:= data_cache_misses		
 
$1F42 =:= data_cache_refill_l2_all	  \ all	
$1042 =:= data_cache_refill_l2_modified	  \ MODIFIED
$0842 =:= data_cache_refill_l2_owner	  \ OWNER
$0442 =:= data_cache_refill_l2_exclusive  \ EXCLUSIVE
$0242 =:= data_cache_refill_l2_shared	  \ SHARED
$0142 =:= data_cache_refill_l2_invalid	  \ INVALID

$1F43 =:= data_mem_refs_all		  \ all
$1043 =:= data_mem_refs_modified	  \ MODIFIED
$0843 =:= data_mem_refs_owner		  \ OWNER
$0443 =:= data_mem_refs_exclusive	  \ EXCLUSIVE
$0243 =:= data_mem_refs_shared		  \ SHARED
$0143 =:= data_mem_refs_invalid		  \ INVALID

$1F44 =:= data_cache_writebacks_all	  \ all
$1044 =:= data_cache_writebacks_modified  \ MODIFIED
$0844 =:= data_cache_writebacks_owner	  \ OWNER
$0444 =:= data_cache_writebacks_exclusive \ EXCLUSIVE
$0244 =:= data_cache_writebacks_shared	  \ SHARED
$0144 =:= data_cache_writebacks_invalid	  \ INVALID

$0045 =:= l1_dltb_misses_l2_dltb_hits
$0046 =:= l1&l2_dtlb_misses
$0047 =:= misaligned_data_references

$0080 =:= ifu_ifetch			  \ Number of instruction fetches, both cacheable and non cacheable.
$0081 =:= ifu_ifetch_miss		  \ Number of instruction fetch misses.
$0084 =:= l1_itlb_misses		  \ Number of L1 ITLB misses.
$0085 =:= l1&l2_itlb_misses		  \ Number of L1 and L2 ITLB misses.

$00C0 =:= retired_instructions		  \ # instructions retired (also exceptions, interrupts & resyncs).
$00C1 =:= retired_ops
$00C2 =:= retired_branches		  \ # branch instructions retired.
$00C3 =:= retired_mispredicted_branches
$00C4 =:= retired_taken_branch_mispredict
$00C5 =:= retired_far_control_transfers
$00C7 =:= retired_resync_branches

$00CD =:= ints_masked
$00CE =:= ints_masked_while_pending
$00CF =:= ints_taken

0 VALUE EVT_OS?		-- count OS events (ring 0)
1 VALUE EVT_USER?	-- count user events (ring 1, 2, 3)
1 VALUE EVT_NUMBERS?	-- count occurrences instead of length of events

: PTEST ( 'evt1 'evt2 'evt3 'evt4 xt -- )
	LOCALS| xt 'evt4 'evt3 'evt2 'evt1 |
	0. 0. 0. 0. DLOCALS| evt0 evt1 evt2 evt3 |
	 0  'evt1 @  EVT_USER? EVT_OS? EVT_NUMBERS? SetEventCounter  0 ReadPerfCtr TO evt0 \ counter 0
	 1  'evt2 @  EVT_USER? EVT_OS? EVT_NUMBERS? SetEventCounter  1 ReadPerfCtr TO evt1 \ counter 1
	 2  'evt3 @  EVT_USER? EVT_OS? EVT_NUMBERS? SetEventCounter  2 ReadPerfCtr TO evt2 \ counter 2
	 3  'evt4 @  EVT_USER? EVT_OS? EVT_NUMBERS? SetEventCounter  3 ReadPerfCtr TO evt3 \ counter 3
	   xt EXECUTE
	 3 ReadPerfCtr evt3 D- 	 
	 2 ReadPerfCtr evt2 D- 	 
	 1 ReadPerfCtr evt1 D- 	 
	 0 ReadPerfCtr evt0 D- 
	 CR 'evt1 .EVT ." : " (n,3)
	 CR 'evt2 .EVT ." : " (n,3)
	 CR 'evt3 .EVT ." : " (n,3)
	 CR 'evt4 .EVT ." : " (n,3) 
	 CR ;P

: TEST-PERFORMANCE ( xt -- )
	LOCAL xt 
	InitMSR $74 <> ABORT" AMD Athlon expected"
	  retired_far_control_transfers	  retired_resync_branches
	  ints_masked_while_pending	  ints_taken			  xt PTEST
	  ints_masked			  data_cache_refill_l2_all	  
	  data_mem_refs_all		  data_cache_writebacks_all	  xt PTEST 
	  l1_dltb_misses_l2_dltb_hits	  l1&l2_dtlb_misses		  
	  misaligned_data_references	  ifu_ifetch			  xt PTEST 
	  ifu_ifetch_miss		  l1_itlb_misses
	  l1&l2_itlb_misses		  retired_instructions		  xt PTEST 
	  retired_ops			  retired_branches		
	  retired_mispredicted_branches	  retired_taken_branch_mispredict xt PTEST
	  data_cache_access		  data_cache_misses	
	  data_cache_refill_l2_modified	  data_cache_refill_l2_owner	  xt PTEST 
	  data_cache_refill_l2_exclusive  data_cache_refill_l2_shared 	
	  data_cache_refill_l2_invalid	  data_mem_refs_modified	  xt PTEST 
	  data_mem_refs_owner		  data_mem_refs_exclusive
	  data_mem_refs_shared		  data_mem_refs_invalid		  xt PTEST 
	  data_cache_writebacks_modified  data_cache_writebacks_owner	
	  data_cache_writebacks_exclusive data_cache_writebacks_shared	  xt PTEST 
	ExitMSR ;

0   VALUE tk/ms	   PRIVATE
0. DVALUE inctime  PRIVATE
0. DVALUE prevtime PRIVATE

: TSC1-CAL	TimeStamp #1000 MS TimeStamp 2SWAP D- #1000 UM/MOD TO tk/ms  DROP ;P
: TSC1-RESET	TSC1-CAL CLEAR inctime ;
: TSC1-START	TimeStamp TO prevtime ;
: TSC1-STOP	TimeStamp prevtime D- +TO inctime ;
: .TSC1		inctime tk/ms UM/MOD NIP n.ELAPSED ;

[DEFINED] testing? 
  [IF]

#40175768 VALUE #istr
	0 VALUE offs
CREATE ape 9 , 8 ,

: test			TIMER-RESET #istr 0 DO  ape offs + @+ XOR  
						7 OR 99 AND 
						7 OR 99 AND 
						7 OR 99 AND 
						7 OR 99 AND 
						7 OR 99 AND 33 + DROP 
					  LOOP  .ELAPSED ;

: DO-TEST ( offs -- )	TO offs ['] test TEST-PERFORMANCE ;

[THEN]

:ABOUT	CR ." UNLOAD-PERFC ( -- ) " ;

		    DEPRIVE
nesting @ 1 = [IF] .ABOUT -perfctrs CR  [THEN]

		             (* End of Source *)